

from ..base import RecDataSet


__all__ = []


#TODO:===============================Context-aware Recommendation===============================
class ContextAwareRecSet(RecDataSet):
    DATATYPE =  "Context"